--
--  Copyright (c) 2016, Facebook, Inc.
--  All rights reserved.
--
--  This source code is licensed under the BSD-style license found in the
--  LICENSE file in the root directory of this source tree. An additional grant
--  of patent rights can be found in the PATENTS file in the same directory.
--
require 'torch'
require 'paths'
require 'optim'
require 'nn'
local DataLoader = require 'dataloader'
local models = require 'models/init'
local Trainer = require 'train'
local opts = require 'opts'
local checkpoints = require 'checkpoints'

torch.setdefaulttensortype('torch.FloatTensor')
torch.setnumthreads(1)

local opt = opts.parse(arg)
torch.manualSeed(opt.manualSeed)
cutorch.manualSeedAll(opt.manualSeed)

-- Load previous checkpoint, if it exists
local checkpoint, optimState = checkpoints.latest(opt)

-- Create model
local model, criterion = models.setup(opt, checkpoint)

-- Data loading
local trainLoader, valLoader, testLoader = DataLoader.create(opt) -- FWD

-- The trainer handles the training loop and evaluation on validation set
local trainer = Trainer(model, criterion, opt, optimState)

if opt.testOnly then
   -- FWD
   local top1Err, top5Err
   if opt.dataset ~= 'caltech101' then
      top1Err, top5Err = trainer:test(0, valLoader)
   else
      local csv = require 'csvigo'

      local outputs, counts, correct
      -- top1Err, top5Err, outputs, counts, correct = trainer:test(0, testLoader)
      top1Err, top5Err, outputs = trainer:test(0, testLoader)

      local header = {'img', 'airplanes', 'Motorbikes', 'Faces_easy', 
                      'watch', 'Leopards', 'bonsai'}

      for i = 1, #outputs do
         local csvf = csv.File('outputs_resnet_' .. (i - 1) .. '.csv', 'w')
         csvf:write(header)
         for j = 1, #outputs[i] do
            csvf:write(outputs[i][j])
         end
         csvf:close()
      end
   end
   -- /FWD
   -- FWD
   -- print(string.format(' * Results top1: %6.3f  top5: %6.3f', top1Err, top5Err))
   print(string.format(' * Results top1: %6.3f', top1Err))
   -- /FWD
   return
end

local startEpoch = checkpoint and checkpoint.epoch + 1 or opt.epochNumber
local bestTop1 = math.huge
local bestTop5 = math.huge
for epoch = startEpoch, opt.nEpochs do
   -- Train for a single epoch
   local trainTop1, trainTop5, trainLoss = trainer:train(epoch, trainLoader)

   -- Run model on validation set
   local testTop1, testTop5 = trainer:test(epoch, valLoader)

   local bestModel = false
   if testTop1 < bestTop1 then
      bestModel = true
      bestTop1 = testTop1
      bestTop5 = testTop5
      -- FWD
      -- print(' * Best model ', testTop1, testTop5)
      print(' * Best model ', testTop1)
      -- /FWD
   end

   checkpoints.save(epoch, model, trainer.optimState, bestModel, opt)
end

-- FWD
-- print(string.format(' * Finished top1: %6.3f  top5: %6.3f', bestTop1, bestTop5))
print(string.format(' * Finished top1: %6.3f', bestTop1))
-- /FWD
